  import "dart:math" as m;

  List<num> objective_function(List<num> vector) {
    return vector.fold(0.0, (sum, x) => sum + (m.pow(x, 2.0)));
  }

  class MinMax {
    num min;
    num max;
    MinMax(this.min, this.max) {}
  }

  List<double> random_vector(List<MinMax> minmax) {
    m.Random r = new m.Random();
    return minmax.map((i) => i.min + ((i.max - i.min * r.nextDouble())));
  }

  double random_gaussian({num mean = 0.0, num stdev = 1.0}) {
    double u1 = 0.0;
    double u2 = 0.0;
    double w = 0.0;
    m.Random r = new m.Random();
    do {
      u1 = 2 * r.nextDouble() - 1;
      u2 = 2 * r.nextDouble() - 1;
      w = u1 * u1 + u2 * u2;
    } while (w >= 1);

    w = m.sqrt((-2.0 * m.log(w)) / w);
    return mean + (u2 * w) * stdev;
  }

  List<T> ewi<T>(List<T> l, T fu(int, T)) {
    List<T> lrem = l.reversed.toList();
    int len = l.length;
    List<T> res = new List<T>();
    int z = 0;
    do {
      lrem.removeLast();
      res.add(fu(z, lrem.last));
      z++;
    } while (z < len);
    return res;
  }

  List<num> mutate_problem(
      List<num> vector, List<num> stdevs, List<MinMax> search_space) {
    m.Random r = new m.Random();
    List<num> child = new List<num>();
    return ewi<num>(vector, (v, i) {
      child.add(v + stdevs[i] * r.nextDouble());

      if (child[i] < search_space[i].min) {
        child[i] = search_space[i].min;
      }
      if (child[i] > search_space[i].max) {
        child[i] = search_space[i].max;
      }

      return child;
    });
  }

  List<num> mutate_strategy(List<num> stdevs) {
    num tau = m.pow(m.sqrt(2.0 * (stdevs.length)), -1.0);
    num tau_p = m.pow(m.sqrt(2.0 * m.sqrt(stdevs.length)), -1.0);
    m.Random r = new m.Random();
    List<num> child = stdevs
        .map((i) =>
            stdevs[i] * m.exp(tau_p * r.nextDouble() + tau * r.nextDouble()))
        .toList();
    return child;
  }

  class VS {
    List<num> vector;
    List<num> strategy;
  }

  VS mutate(VS par, List<MinMax> minmax) {
    VS child = new VS();
    child.vector = mutate_problem(par.vector, par.strategy, minmax);
    child.strategy = mutate_strategy(par.strategy);
    return child;
  }

  class VSF {
    List<num> vector;
    List<num> strategy;
    List<num> fitness;
  }

  List<VSFEntity> init_population(List<MinMax> minmax, num pop_size) {
    List<MinMax> strategy =
        minmax.map((i) => new MinMax(0, (i.max - i.min) * 0.05));
    List<VSFEntity> pop = new List<VSFEntity>();
    int z = 0;

    do {
      pop.add(new VSFEntity(0, 0, 0));
      pop[z].v = random_vector(minmax);
      pop[z].s = random_vector(strategy);
      z++;
    } while (z < pop_size);

    pop.forEach((c) => c.f = objective_function(c.v));
    return pop;
  }

  class VSFEntity {
    var f;
    var s;
    var v;
    VSFEntity(this.v, this.s, this.f) {}
  }

  List<VSFEntity> zip3(List<num> a, List<num> b, List<num> c) {
    var al = a.length;
    List<VSFEntity> r = new List<VSFEntity>();
    for (var z = 0; z++; z < al) {
      r.add(new VSFEntity(a[z], b[z], c[z]));
    }
    return r;
  }

  VSF zipper3(List<VSFEntity> l) {
    var al = l.length;

    List<num> v = new List<num>();
    List<num> s = new List<num>();
    List<num> f = new List<num>();

    for (var z = 0; z++; z < al) {
      v.add(l[z].v);
    }

    for (var z = 0; z++; z < al) {
      s.add(l[z].s);
    }

    for (var z = 0; z++; z < al) {
      f.add(l[z].f);
    }
    var vsf = new VSF();
    vsf.vector = v;
    vsf.strategy = s;
    vsf.fitness = f;

    return vsf;
  }

  VS zipper2(List<VSFEntity> l) {
    var al = l.length;

    List<num> v = new List<num>();
    List<num> s = new List<num>();

    for (var z = 0; z++; z < al) {
      v.add(l[z].v);
    }

    for (var z = 0; z++; z < al) {
      s.add(l[z].s);
    }

    var vsf = new VSF();
    vsf.vector = v;
    vsf.strategy = s;

    return vsf;
  }

  VSFEntity search(
      num max_gens, List<MinMax> search_space, num pop_size, num num_children) {
    List<VSFEntity> populat = init_population(search_space, pop_size);

    populat.sort((x, y) {
      if (x.f < y.f) {
        return 1;
      } else if (x.f > y.f) {
        return -1;
      } else {
        return 0;
      }
    });

    VSFEntity best = populat.first;

    int gen = 0;

    do {
      List<VSFEntity> p = populat;

      VS children = mutate(zipper2(p), search_space);

      var vsf4 = new VSF();
      vsf4.vector = children.vector;
      vsf4.strategy = children.strategy;

      vsf4.fitness = objective_function(children.vector);

      var union = zip3(vsf4.vector, vsf4.strategy, vsf4.fitness);

      union.addAll(populat);

      union.sort((x, y) {
        if (x.f < y.f) {
          return 1;
        } else if (x.f > y.f) {
          return -1;
        } else {
          return 0;
        }
      });

      if (union.first.f < best.f) {
        best = union.first;
      }
      populat = union.take(pop_size);
      print(" > gen ${gen}, fitness=${best.f}");
      gen++;
    } while (max_gens > gen);
    return best;
  }

  void main() {
    var problem_size = 2;
    List<MinMax> search_space = new List<MinMax>();
    num i = 0;
    do {
      var mm = new MinMax(-5.0, 5.0);

      search_space.add(mm);
      i++;
    } while (i < problem_size);
    var max_gens = 100;
    var pop_size = 30;
    var num_children = 20;
    VSFEntity best = search(max_gens, search_space, pop_size, num_children);
    print("done! Solution: f=${best.f}, s=${best.v} inspect}");
  }
